﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections;

namespace ZavrsniRad
{
    class GeneticAlgorithm
    {
        #region varijable

        private double _mutationRate;
        public double MutationRate
        {
            get { return _mutationRate; }
            set { _mutationRate = value; }
        }

        private Random _rand;
        public Random Rand
        {
            get { return _rand; }
            set { _rand = value; }
        }

        private List<int> _chromosomeDesign;
        public List<int> ChromosomeDesign
        {
            get { return _chromosomeDesign; }
            set { _chromosomeDesign = value; }
        }

        private int _populationNumber;
        public int PopulationNumber
        {
            get { return _populationNumber; }
            set { _populationNumber = value; }
        }

        private int _geneLength;
        public int GeneLength
        {
            get { return _geneLength; }
            set { _geneLength = value; }
        }

        #endregion
        /// <summary>
        ///  Broj jedinki, dizajn kromosoma, duljina pojedinog gena i postotak mutacije...
        /// </summary>
        /// <param name="populationNumber"></param>
        /// <param name="chromosomeDesign"></param>
        /// <param name="geneLength"></param>
        /// <param name="mutationRate"></param>
        public GeneticAlgorithm(int populationNumber, List<int> chromosomeDesign, int geneLength, double mutationRate)
        {
            this.MutationRate = mutationRate;
            this.Rand = new Random();
            this.GeneLength = geneLength;
            this.ChromosomeDesign = chromosomeDesign;
            this.PopulationNumber = populationNumber;
        }
        /// <summary>
        /// Metoda stvara neuronsku mrežu sačinjenu od BitArray
        /// </summary>
        /// <param name="firstParent"></param>
        /// <param name="secondParent"></param>
        /// <returns></returns>
        public List<BitArray[][]> ChildFromParents(List<BitArray[][]> firstParent, List<BitArray[][]> secondParent)
        {
            List<BitArray[][]> result = new List<BitArray[][]>();
            Random rand = new Random();
            for (int i = 0; i < _chromosomeDesign.Count() - 1; i++)
            {
                BitArray[][] firstParentMatrix = firstParent[i];
                BitArray[][] secondParentMatrix = secondParent[i];
                BitArray[][] matrixToCreate = new BitArray[_chromosomeDesign[i]][];
                for (int j = 0; j < this.ChromosomeDesign[i]; j++)
                {
                    matrixToCreate[j] = new BitArray[_chromosomeDesign[i + 1]];
                    for (int k = 0; k < _chromosomeDesign[i + 1]; k++)
                    {
                        //radimo uniformni crossover - CrossoverBitArrays i nakon toga mutaciju - MutateBitArray
                        matrixToCreate[j][k] = MutateBitArray(CrossoverBitArrays(firstParentMatrix[j][k], secondParentMatrix[j][k]));
                    }
                }
                result.Add(matrixToCreate);
            }
            return result;
        }
        /// <summary>
        /// Metoda stvara listu matrica sačinjenih od nasumično stvorenih BitArray varijabli
        /// </summary>
        /// <param name="geneLength"></param>
        /// <returns></returns>
        public List<BitArray[][]> RandomBitArrayChromosome()
        {            
            List<BitArray[][]> result = new List<BitArray[][]>();
            for (int i = 0; i < _chromosomeDesign.Count() - 1; i++)
            {
                BitArray[][] matrixToAdd = new BitArray[_chromosomeDesign[i]][];

                for (int j = 0; j < _chromosomeDesign[i]; j++)
                {
                    matrixToAdd[j] = new BitArray[_chromosomeDesign[i + 1]];

                    for (int k = 0; k < _chromosomeDesign[i + 1]; k++)
                    {
                        matrixToAdd[j][k] = RandomBitArray();
                    }
                }
                result.Add(matrixToAdd);
            }
            return result;             
        }
        /// <summary>
        /// Vraća BitArray koji se sastoji od nasumično odabranih bitova 2 ulazna BitArray-a, tj. izvodi se uniformno križanje
        /// </summary>
        /// <param name="first"></param>
        /// <param name="second"></param>
        /// <param name="rand"></param>
        /// <returns></returns>
        private BitArray CrossoverBitArrays(BitArray first, BitArray second)
        {
            for (int i = 0; i < first.Count; i++)
            {
                if (this.Rand.NextDouble() < 0.5)
                { first.Set(i, false); }
                else { second.Set(i, false); }
            }
            return first.Or(second);
        }
        /// <summary>
        /// Izvodi crossover s jednom točkom prekida.
        /// </summary>
        /// <param name="first"></param>
        /// <param name="second"></param>
        /// <returns></returns>
        private BitArray PointCrossoverBitArrays(BitArray first, BitArray second)
        {
            BitArray result = new BitArray(first.Count, false);
            int point = this.Rand.Next(first.Count);
            for (int i = 0; i < first.Count; i++)
            {
                if (i <= point)
                {
                    result[i] = second[i];
                }
                else
                {
                    result[i] = first[i];
                }
            }
            return result;
        }
        /// <summary>
        /// Vraća BitArrays invertiranim pojedinim bitovima ulaznog BitArray-a
        /// Vjerojatnost inverzije je zadana s mutationRate [0.0 - 1.0]
        /// </summary>
        /// <param name="value"></param>
        /// <param name="mutationRate"></param>
        /// <param name="rand"></param>
        /// <returns></returns>
        private BitArray MutateBitArray(BitArray value)
        {
            for (int i = 0; i < value.Count; i++)
            {
                if (this.Rand.NextDouble() < this.MutationRate)
                {
                    if (value[i] == true)
                    { value.Set(i, false); }
                    else { value.Set(i, true); }
                }
            }
            return value;
        }
        /// <summary>
        /// Stvara novi BitArray dužine length s nasumičnim vrijednostima
        /// </summary>
        /// <param name="length"></param>
        /// <param name="rand"></param>
        /// <returns></returns>
        private BitArray RandomBitArray()
        {
            BitArray result = new BitArray(this.GeneLength, false);
            for (int i = 0; i < this.GeneLength; i++)
            {
                if (this.Rand.NextDouble() < 0.5) result[i] = true;
            }
            return result;
        }
    }   
}
